"""
Utilities used across many GA4GH packages
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import StringIO
import contextlib
import fnmatch
import functools
import humanize
import itertools
import os
import shlex
import signal
import subprocess
import sys
import time
import yaml


def log(message):
    """
    Log a message
    """
    print(message)


def getPathOfExecutable(executable):
    """
    Returns the full path of the executable, or None if the executable
    can not be found.
    """
    exe_paths = os.environ['PATH'].split(':')
    for exe_path in exe_paths:
        exe_file = os.path.join(exe_path, executable)
        if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
            return exe_file
    return None


def requireExecutables(executables):
    """
    Check that all of the given executables are on the path.
    If at least one of them is not, exit the script and inform
    the user of the missing requirement(s).
    """
    missingExecutables = []
    for executable in executables:
        if getPathOfExecutable(executable) is None:
            missingExecutables.append(executable)
    if len(missingExecutables) > 0:
        log("In order to run this script, the following "
            "executables need to be on the path:")
        for missingExecutable in missingExecutables:
            log(missingExecutable)
        exit(1)


def runCommand(command, silent=False, shell=False):
    """
    Run a shell command
    """
    splits = shlex.split(command)
    runCommandSplits(splits, silent=silent, shell=shell)


def runCommandReturnOutput(cmd):
    """
    Runs a shell command and return the stdout and stderr
    """
    splits = shlex.split(cmd)
    proc = subprocess.Popen(
        splits, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        raise subprocess.CalledProcessError(stdout, stderr)
    return stdout, stderr


def runCommandSplits(splits, silent=False, shell=False):
    """
    Run a shell command given the command's parsed command line
    """
    try:
        if silent:
            with open(os.devnull, 'w') as devnull:
                subprocess.check_call(
                    splits, stdout=devnull, stderr=devnull, shell=shell)
        else:
            subprocess.check_call(splits, shell=shell)
    except OSError as exception:
        if exception.errno == 2:  # cmd not found
            raise Exception(
                "Can't find command while trying to run {}".format(splits))
        else:
            raise


def getAuthValues(filePath='scripts/auth.yml'):
    """
    Return the script authentication file as a dictionary
    """
    return getYamlDocument(filePath)


def getYamlDocument(filePath):
    """
    Return a yaml file's contents as a dictionary
    """
    with open(filePath) as stream:
        doc = yaml.load(stream)
        return doc


def captureOutput(func, *args, **kwargs):
    """
    Runs the specified function and arguments, and returns the
    tuple (stdout, stderr) as strings.
    """
    stdout = sys.stdout
    sys.stdout = StringIO.StringIO()
    stderr = sys.stderr
    sys.stderr = StringIO.StringIO()
    try:
        func(*args, **kwargs)
        stdoutOutput = sys.stdout.getvalue()
        stderrOutput = sys.stderr.getvalue()
    finally:
        sys.stdout.close()
        sys.stdout = stdout
        sys.stderr.close()
        sys.stderr = stderr
    return stdoutOutput, stderrOutput


def zipLists(*lists):
    """
    Checks to see if all of the lists are the same length, and throws
    an AssertionError otherwise.  Returns the zipped lists.
    """
    length = len(lists[0])
    for i, list_ in enumerate(lists[1:]):
        if len(list_) != length:
            msg = "List at index {} has length {} != {}".format(
                i + 1, len(list_), length)
            raise AssertionError(msg)
    return zip(*lists)


def getLinesFromLogFile(stream):
    """
    Returns all lines written to the passed in stream
    """
    stream.flush()
    stream.seek(0)
    lines = stream.readlines()
    return lines


def powerset(iterable, maxSets=None):
    """
    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)

    See https://docs.python.org/2/library/itertools.html#recipes
    """
    s = list(iterable)
    return itertools.islice(itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)),
        0, maxSets)


def chomp(line):
    """
    Returns a string stripped of its trailing newline character
    """
    assert line[-1] == '\n'
    return line[:-1]


def getFilePathsWithExtensionsInDirectory(dirTree, patterns, sort=True):
    """
    Returns all file paths that match any one of patterns in a
    file tree with its root at dirTree.  Sorts the paths by default.
    """
    filePaths = []
    for root, dirs, files in os.walk(dirTree):
        for filePath in files:
            for pattern in patterns:
                if fnmatch.fnmatch(filePath, pattern):
                    fullPath = os.path.join(root, filePath)
                    filePaths.append(fullPath)
                    break
    if sort:
        filePaths.sort()
    return filePaths


def touch(filepath):
    """
    Creates an empty file at filepath, if it does not already exist
    """
    with open(filepath, 'a'):
        pass


def assertFileContentsIdentical(pathOne, pathTwo):
    with open(pathOne) as fileOne, open(pathTwo) as fileTwo:
        for i, (lineOne, lineTwo) in enumerate(zip(fileOne, fileTwo)):
            if lineOne != lineTwo:
                msg = "Mismatch on line {}: '{}' != '{}'".format(
                    i + 1, lineOne, lineTwo)
                raise AssertionError(msg)
        # one file could have more lines than the other file
        extraLine = fileOne.readline()
        if extraLine:
            msg = "File one has extra line: '{}'".format(extraLine)
            raise AssertionError(msg)
        extraLine = fileTwo.readline()
        if extraLine:
            msg = "File two has extra line: '{}'".format(extraLine)
            raise AssertionError(msg)


# ---------------- Decorators ----------------


class TimeoutException(Exception):
    """
    A process has taken too long to execute
    """


class Timed(object):
    """
    Decorator that times a method, reporting runtime at finish
    """
    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.start = time.time()
            result = func(*args, **kwargs)
            self.end = time.time()
            self._report()
            return result
        return wrapper

    def _report(self):
        delta = self.end - self.start
        timeString = humanize.time.naturaldelta(delta)
        log("Finished in {} ({:.2f} seconds)".format(timeString, delta))


class Repeat(object):
    """
    A decorator to use for repeating a tagged function.
    The tagged function should return true if it wants to run again,
    and false if it wants to stop repeating.
    """
    defaultSleepSeconds = 0.1

    def __init__(self, sleepSeconds=defaultSleepSeconds):
        self.sleepSeconds = sleepSeconds

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            while func(*args, **kwargs):
                time.sleep(self.sleepSeconds)
        return wrapper


class Timeout(object):
    """
    A decorator to use for only allowing a function to run
    for a limited amount of time
    """
    defaultTimeoutSeconds = 60

    def __init__(self, timeoutSeconds=defaultTimeoutSeconds):
        self.timeoutSeconds = timeoutSeconds

    def __call__(self, func):

        def _handle_timeout(signum, frame):
            raise TimeoutException()

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                # set the alarm and execute func
                signal.signal(signal.SIGALRM, _handle_timeout)
                signal.alarm(self.timeoutSeconds)
                result = func(*args, **kwargs)
            finally:
                # clear the alarm
                signal.alarm(0)
            return result
        return wrapper


# ---------------- Context managers ----------------


@contextlib.contextmanager
def suppressOutput():
    # I would like to use sys.stdout.fileno() and sys.stderr.fileno()
    # here instead of literal fd numbers, but nose does something like
    # sys.stdout = StringIO.StringIO() when the -s flag is not enabled
    # (to capture test output so it doesn't get entangled with nose's
    # display) so the sys.stdout and sys.stderr objects are not able to
    # be accessed here.
    try:
        # redirect stdout and stderr to /dev/null
        devnull = open(os.devnull, 'w')
        origStdoutFd = 1
        origStderrFd = 2
        origStdout = os.dup(origStdoutFd)
        origStderr = os.dup(origStderrFd)
        sys.stdout.flush()
        sys.stderr.flush()
        os.dup2(devnull.fileno(), origStdoutFd)
        os.dup2(devnull.fileno(), origStderrFd)
        # enter the wrapped code
        yield
    finally:
        # restore original file streams
        devnull.flush()
        os.dup2(origStdout, origStdoutFd)
        os.dup2(origStderr, origStderrFd)
        # clean up
        os.close(origStdout)
        os.close(origStderr)
        devnull.close()


@contextlib.contextmanager
def performInDirectory(dirPath):
    """
    Change the current working directory to dirPath before performing
    an operation, then restore the original working directory after
    """
    originalDirectoryPath = os.getcwd()
    try:
        os.chdir(dirPath)
        yield
    finally:
        os.chdir(originalDirectoryPath)



"""
Common cli functionality
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import operator


class SortedHelpFormatter(argparse.HelpFormatter):
    """
    An argparse HelpFormatter that sorts the flags and subcommands
    in alphabetical order
    """
    def add_arguments(self, actions):
        """
        Sort the flags alphabetically
        """
        actions = sorted(
            actions, key=operator.attrgetter('option_strings'))
        super(SortedHelpFormatter, self).add_arguments(actions)

    def _iter_indented_subactions(self, action):
        """
        Sort the subcommands alphabetically
        """
        try:
            get_subactions = action._get_subactions
        except AttributeError:
            pass
        else:
            self._indent()
            if isinstance(action, argparse._SubParsersAction):
                for subaction in sorted(
                        get_subactions(), key=lambda x: x.dest):
                    yield subaction
            else:
                for subaction in get_subactions():
                    yield subaction
            self._dedent()


def addSubparser(subparsers, subcommand, description):
    """
    Add a subparser with subcommand to the subparsers object
    """
    parser = subparsers.add_parser(
        subcommand, description=description, help=description)
    return parser


def createArgumentParser(description):
    """
    Create an argument parser
    """
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=SortedHelpFormatter)
    return parser



"""
Tests for common utilities
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import mock
import tempfile
import subprocess
import unittest
import sys
import os

import ga4gh.common.utils as utils


class AbstractTestUtils(unittest.TestCase):
    """
    Base class for tests that contains common attributes
    """
    executables = ['ls', 'rm']
    nonexistentExecutable = 'doesNotExistAsAnExecutable'
    validCommand = 'echo -n'
    invalidCommand = 'ls -5'


class TestUtils(AbstractTestUtils):

    def testGetPathOfExecutable(self):
        path = utils.getPathOfExecutable(self.executables[0])
        self.assertIsNotNone(path)
        notFoundPath = utils.getPathOfExecutable(self.nonexistentExecutable)
        self.assertIsNone(notFoundPath)

    def testRunCommand(self):
        utils.runCommand(self.validCommand)
        with self.assertRaises(subprocess.CalledProcessError):
            utils.runCommand(self.invalidCommand, silent=True)
        with self.assertRaises(Exception):
            utils.runCommand(self.nonexistentExecutable, silent=True)

    def testRunCommandReturnOutput(self):
        output = utils.runCommandReturnOutput(self.validCommand)
        self.assertIsNotNone(output)
        with self.assertRaises(subprocess.CalledProcessError):
            utils.runCommandReturnOutput(self.invalidCommand)
        with self.assertRaises(Exception):
            utils.runCommandReturnOutput(self.nonexistentExecutable)

    def testRunCommandSplits(self):
        utils.runCommandSplits(self.validCommand.split())
        with self.assertRaises(subprocess.CalledProcessError):
            utils.runCommandSplits(self.invalidCommand.split(), silent=True)
        with self.assertRaises(Exception):
            utils.runCommand([self.nonexistentExecutable], silent=True)

    def testGetYamlDocument(self):
        yamlText = """
provider:
        key: aKey
        """
        handle, path = tempfile.mkstemp()
        with open(path, 'w') as yamlFile:
            yamlFile.write(yamlText)
        doc = utils.getYamlDocument(path)
        self.assertEqual(doc, {'provider': {'key': 'aKey'}})

    def testZipLists(self):
        a = [1, 2]
        b = [3, 4]
        c = [5, 6]
        d = [7, 8, 9]
        e = [10]
        result = utils.zipLists(a, b, c)
        self.assertEqual(result, [(1, 3, 5), (2, 4, 6)])
        with self.assertRaises(AssertionError):
            utils.zipLists(c, d)
        with self.assertRaises(AssertionError):
            utils.zipLists(d, e)

    def testGetLinesFromLogFile(self):
        handle, path = tempfile.mkstemp()
        message = 'aMessage'
        with open(path, 'w') as yamlFile:
            yamlFile.write(message)
        with open(path) as readFile:
            result = utils.getLinesFromLogFile(readFile)
        self.assertEqual(result, [message])

    def testPowerset(self):
        s = [1, 2, 3]
        expected = [
            (), (1, ), (2, ), (3, ), (1, 2), (1, 3), (2, 3), (1, 2, 3)]
        result = list(utils.powerset(s))
        self.assertEqual(result, expected)

    def testRepeat(self):
        aList = []

        @utils.Repeat()
        def repeatFunc():
            aList.append(1)
            return len(aList) < 3
        repeatFunc()
        self.assertEqual(aList, [1, 1, 1])

    def testTimeout(self):

        @utils.Timeout(1)
        def timeoutFunc():
            while True:
                pass
        with self.assertRaises(utils.TimeoutException):
            timeoutFunc()

    def testSuppressOutput(self):
        # not really sure how to test this
        with utils.suppressOutput():
            pass

    def testPerformInDirectory(self):
        originalDirPath = os.getcwd()
        with utils.performInDirectory('..'):
            newDirPath = os.getcwd()
        self.assertEqual(os.getcwd(), originalDirPath)
        self.assertNotEqual(originalDirPath, newDirPath)

    def testChomp(self):
        line = 'line\n'
        chomped = utils.chomp(line)
        self.assertEqual(chomped, 'line')
        with self.assertRaises(AssertionError):
            utils.chomp(chomped)

    def testGetFilePathsWithExtensionsInDirectory(self):
        def createTempTree():
            tree = tempfile.mkdtemp('getFilePathsWithExtensionsInDirectory')
            utils.touch(os.path.join(tree, 'a.foo'))
            utils.touch(os.path.join(tree, 'b.bar'))
            utils.touch(os.path.join(tree, 'c.baz'))
            utils.touch(os.path.join(tree, 'd.foo'))
            treeDirPath = os.path.join(tree, 'treeDir')
            os.mkdir(treeDirPath)
            utils.touch(os.path.join(treeDirPath, 'e.foo'))
            utils.touch(os.path.join(treeDirPath, 'f.bar'))
            utils.touch(os.path.join(treeDirPath, 'g.baz'))
            utils.touch(os.path.join(treeDirPath, 'h.foo'))
            return tree

        tree = createTempTree()
        patterns = ['*.foo', '*.bar', '*.fo*']
        filePaths = utils.getFilePathsWithExtensionsInDirectory(
            tree, patterns, sort=False)
        self.assertEqual(len(filePaths), 6)
        basenames = [os.path.basename(filePath) for filePath in filePaths]
        self.assertIn('a.foo', basenames)
        self.assertIn('b.bar', basenames)
        self.assertIn('d.foo', basenames)
        self.assertIn('e.foo', basenames)
        self.assertIn('f.bar', basenames)
        self.assertIn('h.foo', basenames)

        filePaths.sort()
        sortedFilePaths = utils.getFilePathsWithExtensionsInDirectory(
            tree, patterns)
        self.assertEqual(filePaths, sortedFilePaths)

    def testTouch(self):
        tree = tempfile.mkdtemp('testTouch')
        filePath = os.path.join(tree, 'touch.txt')
        self.assertFalse(os.path.exists(filePath))
        utils.touch(filePath)
        self.assertTrue(os.path.exists(filePath))
        utils.touch(filePath)
        self.assertTrue(os.path.exists(filePath))

    def testAssertFileContentsIdentical(self):
        _, pathOne = tempfile.mkstemp()
        with open(pathOne, 'w') as fileOne:
            fileOne.write('a')
        _, pathTwo = tempfile.mkstemp()
        with open(pathTwo, 'w') as fileTwo:
            fileTwo.write('a')
        _, pathThree = tempfile.mkstemp()
        with open(pathThree, 'w') as fileThree:
            fileThree.write('b')
        _, pathFour = tempfile.mkstemp()
        with open(pathFour, 'w') as fileFour:
            fileFour.write('a\nb')
        # contents identical
        utils.assertFileContentsIdentical(pathOne, pathTwo)
        # contents not identical (line diff)
        with self.assertRaises(AssertionError):
            utils.assertFileContentsIdentical(pathOne, pathThree)
        # contents not identical (one file larger)
        with self.assertRaises(AssertionError):
            utils.assertFileContentsIdentical(pathOne, pathFour)
        # contents not identical (one file larger, args reversed)
        with self.assertRaises(AssertionError):
            utils.assertFileContentsIdentical(pathFour, pathOne)


class TestUtilsPrintMocked(AbstractTestUtils):
    """
    Tests that need to have print output suppressed
    """
    printMock = mock.Mock()

    def setUp(self):
        self.printMock.reset_mock()

    @mock.patch('__builtin__.print', printMock)
    def testTimed(self):
        aList = []

        @utils.Timed()
        def timedFunc():
            aList.append(1)
        timedFunc()
        self.assertEquals(aList, [1])
        self.assertEquals(self.printMock.call_count, 1)

    @mock.patch('__builtin__.print', printMock)
    def testLog(self):
        utils.log("message")
        self.assertEquals(self.printMock.call_count, 1)

    @mock.patch('__builtin__.print', printMock)
    def testRequireExecutables(self):
        utils.requireExecutables(self.executables)
        with self.assertRaises(SystemExit):
            utils.requireExecutables([self.nonexistentExecutable])
        self.assertEquals(self.printMock.call_count, 2)


class TestCaptureOutput(unittest.TestCase):
    """
    Test that the captureOutput correctly returns the value of stdout
    and stderr for a function.
    """

    def testCapture(self):
        stdoutValue = "stdout"
        stderrValue = "stderr"

        def func():
            print(stdoutValue, file=sys.stdout, end="")
            print(stderrValue, file=sys.stderr, end="")
        stdout, stderr = utils.captureOutput(func)
        self.assertEqual(stdout, stdoutValue)
        self.assertEqual(stderr, stderrValue)

        # Empty stdout
        def func():
            print(stderrValue, file=sys.stderr, end="")
        stdout, stderr = utils.captureOutput(func)
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, stderrValue)

        # Empty stderr
        def func():
            print(stdoutValue, file=sys.stdout, end="")
        stdout, stderr = utils.captureOutput(func)
        self.assertEqual(stdout, stdoutValue)
        self.assertEqual(stderr, "")

    def testArgs(self):
        def func(one, two, three, keywordOne=None, keywordTwo=None):
            print(one, two, three, keywordOne, keywordTwo, file=sys.stdout)
        stdout, stderr = utils.captureOutput(
            func, "1", "2", "3", keywordTwo="5", keywordOne="4")
        self.assertEqual(stdout, "1 2 3 4 5\n")
        self.assertEqual(stderr, "")



"""
Common utilities for GA4GH software
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__import__('pkg_resources').declare_namespace(__name__)



"""
Tests for common cli functionality
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.common.cli as cli


class TestCli(unittest.TestCase):

    def testArgumentParser(self):
        parser = cli.createArgumentParser("test parser")
        subparsers = parser.add_subparsers(title='subparsers')
        subparserName = "test-subparser"
        subparser = cli.addSubparser(
            subparsers, subparserName, "test subparser")
        argumentName = 'argument'
        subparser.add_argument(argumentName)
        parser.parse_args([subparserName, argumentName])



"""
Emulates a Travis CI run
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.common
import ga4gh.common.cli as cli
import ga4gh.common.utils as utils


class TravisSimulator(object):

    logStrPrefix = '***'
    yamlFileLocation = '.travis.yml'

    def parseTestCommands(self):
        yamlData = utils.getYamlDocument(self.yamlFileLocation)
        return yamlData['script']

    def runTests(self):
        testCommands = self.parseTestCommands()
        for command in testCommands:
            self.log('Running: "{}"'.format(command))
            utils.runCommand(command)
        self.log('SUCCESS')

    def log(self, logStr):
        utils.log("{0} {1}".format(self.logStrPrefix, logStr))


def run_tests_main():
    parser = cli.createArgumentParser("runs tests for a ga4gh package")
    versionString = "GA4GH Runtests Version {}".format(
        ga4gh.common.__version__)
    parser.add_argument(
        "--version", version=versionString, action="version")
    parser.parse_args()

    travisSimulator = TravisSimulator()
    travisSimulator.runTests()



# Don't import __future__ packages here; they make setup fail

# First, we try to use setuptools. If it's not available locally,
# we fall back on ez_setup.
try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

with open("README.pypi.rst") as readmeFile:
    long_description = readmeFile.read()

install_requires = []
with open("requirements.txt") as requirementsFile:
    for line in requirementsFile:
        line = line.strip()
        if len(line) == 0:
            continue
        if line[0] == '#':
            continue
        pinnedVersion = line.split()[0]
        install_requires.append(pinnedVersion)

setup(
    # END BOILERPLATE
    name="ga4gh_common",
    description="Common utilities for GA4GH packages",
    packages=["ga4gh", "ga4gh.common"],
    namespace_packages=["ga4gh"],
    url="https://github.com/ga4gh/ga4gh-common",
    use_scm_version={"write_to": "ga4gh/common/_version.py"},
    entry_points={
        "console_scripts": [
            "ga4gh_run_tests=ga4gh.common.run_tests:run_tests_main",
        ],
    },
    # BEGIN BOILERPLATE
    long_description=long_description,
    install_requires=install_requires,
    license='Apache License 2.0',
    include_package_data=True,
    zip_safe=True,
    author="Global Alliance for Genomics and Health",
    author_email="theglobalalliance@genomicsandhealth.org",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    keywords=['genomics', 'reference'],
    # Use setuptools_scm to set the version number automatically from Git
    setup_requires=['setuptools_scm'],
)



"""
Shim for running the run_tests program during development
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.common.run_tests as run_tests


if __name__ == '__main__':
    run_tests.run_tests_main()
